/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: metronom.c,v 1.6 2000/11/11 16:40:34 guenter Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "metronom.h"
#include <inttypes.h>
#include "monitor.h"
#include <stdio.h>

#define PTS_TAB_NUM 50

typedef struct metronom_priv_s {

  uint32_t        pts_per_frame;
  uint32_t        pts_per_smpls;

  int32_t         video_pts_delta;
  int32_t         audio_pts_delta;

  uint32_t        video_vpts;
  uint32_t        audio_vpts;

  uint32_t        vpts_offset;

  int             num_video_vpts_guessed;
  int             num_audio_samples_guessed;

} metronom_priv_t;

metronom_priv_t    gMetronom;


void metronom_reset () {

  gMetronom.video_vpts             = 0;
  gMetronom.audio_vpts             = 0;

  gMetronom.vpts_offset            = 0;

  gMetronom.video_pts_delta        = 0;
  gMetronom.audio_pts_delta        = 0;

  gMetronom.num_video_vpts_guessed    = 0;
  gMetronom.num_audio_samples_guessed = 0;

}

void metronom_set_video_rate (uint32_t pts_per_frame) {
  gMetronom.pts_per_frame = pts_per_frame;
}

void metronom_set_audio_rate (uint32_t pts_per_smpls) {
  gMetronom.pts_per_smpls = pts_per_smpls;
}

uint32_t metronom_got_video_frame (uint32_t pts) {

  uint32_t vpts;

  if (pts) {
    int32_t diff;

    vpts = pts + gMetronom.vpts_offset;

    diff = (int32_t) vpts - (int32_t) gMetronom.video_vpts;

    debug_printf (3,"metronom: video diff : %d\n", diff);

    /* if (gMetronom.num_video_vpts_guessed) */
    gMetronom.video_pts_delta += diff / (gMetronom.num_video_vpts_guessed+1);

    debug_printf (3,"metronom: video => new delta : %d (%d guessed)\n", gMetronom.video_pts_delta,
		  gMetronom.num_video_vpts_guessed);

    if (abs(gMetronom.video_pts_delta) >= 1600) {

      debug_printf (3, "metronom: video delta too bad => SET\n");

      gMetronom.video_pts_delta = 0;

    } 

    if (diff < -1600) {
	
      gMetronom.vpts_offset = gMetronom.video_vpts - pts;
      vpts = pts + gMetronom.vpts_offset;
      
      debug_printf (3, "metronom: video offset = %d\n",gMetronom.vpts_offset);
    } 

    gMetronom.video_vpts = vpts;

    gMetronom.num_video_vpts_guessed = 0;

  } else {

    vpts = gMetronom.video_vpts;
    gMetronom.num_video_vpts_guessed++ ;
  }
  
  gMetronom.video_vpts += gMetronom.pts_per_frame + gMetronom.video_pts_delta;

  debug_printf (3, "metronom: video vpts for %d : %d\n", pts, vpts);

  return vpts;
}


uint32_t metronom_got_audio_samples (uint32_t pts, uint32_t nsamples) {

  uint32_t vpts;

  if (pts) {
    int32_t diff;

    vpts = pts + gMetronom.vpts_offset;

    diff = (int32_t) vpts - (int32_t) gMetronom.audio_vpts;

    debug_printf (3, "metronom: audio diff : %d\n", diff);

    
    if (abs(diff) < gMetronom.pts_per_frame) {

      if (gMetronom.num_audio_samples_guessed)
	gMetronom.audio_pts_delta += diff * 65536 / gMetronom.num_audio_samples_guessed;
      
    } else {
      gMetronom.audio_pts_delta = 0;

    }
    


    if (diff < 0) {
      
      gMetronom.vpts_offset = gMetronom.audio_vpts - pts;
      vpts = pts + gMetronom.vpts_offset;
    } 

    gMetronom.audio_vpts = vpts;

    gMetronom.num_audio_samples_guessed = 0;

  } else {

    vpts = gMetronom.audio_vpts;
    gMetronom.num_audio_samples_guessed += nsamples ;
  }
  
  gMetronom.audio_vpts += nsamples * gMetronom.pts_per_smpls / 65536;

  debug_printf (3, "metronom: audio vpts for %d : %d\n", pts, vpts);

  return vpts;
}


