/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: xine.c,v 1.39 2000/11/19 16:24:01 guenter Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <dirent.h>
#include <dlfcn.h>
#include <Imlib.h>

#include "xine.h"
#include "audio_out/audio_out.h"
#include "video_out/video_out.h"
#include "demux.h"
#include "buffer.h"
#include "libac3/ac3.h"
#include "libmpg123/mpglib.h"
#include "libmpeg2/mpeg2.h"
#include "input/input_plugin.h"
#include "metronom.h"
#include "configfile.h"
#include "monitor.h"


extern const char *ao_available[];
extern int ao_use_driver;
/*
 * top-level xine functions
 */

Display        *gDisplay;
ImlibData      *gImlib_data;
Window          gVideoWin;
Pixmap          gXineLogo;
int             gXineLogoWidth, gXineLogoHeight;
ao_functions_t *gAudioOut;
struct mpstr    gMpgAudio;

/* debugging purposes */

/*
int             gTestFile=-1;
*/

typedef struct _xine_globals {
  input_plugin_t             mInputPlugins[INPUT_PLUGIN_MAX];
  int                        mNumInputPlugins;
  input_plugin_t            *mCurInput;

  demux_functions_t         *mDemuxer[10];
  int                        mNumDemuxer;
  demux_functions_t         *mCurDemux;

  int                        mnStatus;
  uint32_t                   mnPausePos;
  char                       mCurMRL[1024];
  pthread_t                  mThread;
  fifo_buffer_t             *mBufAudio;
  fifo_buffer_t             *mBufVideo;
  struct mpstr               mpgaudio;
  int                        mpgaudio_initialized;

  int                        mnAudioChannel;
  
  gui_status_callback_func_t mStatusCallback;
} xine_globals ;

static xine_globals gXine;

static void xine_decode_audio () {

  static char    mpgaudio_out[16384];
  static int     size, ret;
  buf_element_t *pBuf;

  pBuf = fifo_buffer_get (gXine.mBufAudio);

  switch (pBuf->nType) {
  case BUF_RESET:
    if (gXine.mpgaudio_initialized) {
      exitMPGAUDIO (&gXine.mpgaudio);
      gXine.mpgaudio_initialized = 0;
    }
    ac3_reset ();
    gAudioOut->close ();

    break;
  case BUF_AC3AUDIO:
    
    /* write (gTestFile, pBuf->pContent, pBuf->nSize);  */
    ac3_decode_data(pBuf->pContent, pBuf->pContent + pBuf->nSize,
		    pBuf->nPTS);
    
    break;
    
  case BUF_MPEGAUDIO:

    if (!gXine.mpgaudio_initialized) {
      initMPGAUDIO (&gXine.mpgaudio);
      gXine.mpgaudio_initialized = 1;
    }

    ret = decodeMPGAUDIO (&gXine.mpgaudio, pBuf->pContent, pBuf->nSize, 
			  mpgaudio_out, 16384, &size);
    
    while (ret == MP3_OK) {
      /*
       * init device (audio driver checks if it has been initialized already
       */
      
      (gAudioOut->open) (16, gXine.mpgaudio.fr.sample_rate, gXine.mpgaudio.fr.stereo);
    
      (gAudioOut->put_samples) ((int16_t *) mpgaudio_out,  size/2, pBuf->nPTS);
    
      pBuf->nPTS = 0;

      ret = decodeMPGAUDIO (&gXine.mpgaudio, NULL, 0, 
			    mpgaudio_out, 16384, &size);
    }
    break;
  }

  buffer_pool_free (pBuf);
}

static void xine_decode_video () {

  buf_element_t *pBuf;

  pBuf = fifo_buffer_get (gXine.mBufVideo);

  switch (pBuf->nType) {
  case BUF_RESET:
    mpeg2_reset ();
    vo_reset();
    break;
  case BUF_MPEG2VIDEO:
  case BUF_MPEG1VIDEO:

    mpeg2_decode_data(pBuf->pContent, pBuf->pContent + pBuf->nSize, 
		      pBuf->nPTS);

    break;
  }

  buffer_pool_free (pBuf);
}

void *xine_loop (void *dummy) {

  int    nCount = 0;
  int    audio_buf_status;
  int    video_buf_full;

  /*
   * precache
   */

  /*
  while ( (gXine.mCurDemux->demux_get_status () != DEMUX_FINISHED)
	  && ( ( (!fifo_buffer_is_empty (gXine.mBufAudio)) 
		 && (gAudioOut->get_buf_low ()>1) )
	       || (!vo_is_buffer_full () 
		   && !fifo_buffer_is_empty (gXine.mBufVideo)) )) {

    printf ("precaching...\n");

    if ( (!fifo_buffer_is_empty (gXine.mBufAudio)) 
	 && (gAudioOut->get_buf_low ()>1) ) {
      printf ("precaching audio\n");
      xine_decode_audio ();
    }

    if (!vo_is_buffer_full () && !fifo_buffer_is_empty (gXine.mBufVideo)) { 
      printf ("precaching video\n");
      xine_decode_video ();
    } else {
      if (vo_is_buffer_full()) {
	printf ("video buffer is full!\n");
      }
      if (fifo_buffer_is_empty (gXine.mBufVideo))
	printf ("video buffer empty\n");
    }
  }
  */
  
  
  while (!vo_is_buffer_full () && !fifo_buffer_is_empty (gXine.mBufVideo)) { 
    xine_decode_video ();
  } 


  while (gXine.mnStatus == XINE_PLAY) {
    
    if (nCount > 200) { /* Hack: don't do this too often */
      
      gXine.mStatusCallback (gXine.mnStatus);
      nCount = 0;
    }

    nCount++;


    if ((gXine.mCurDemux->demux_get_status () == DEMUX_FINISHED) 
	&& (fifo_buffer_is_empty (gXine.mBufAudio) 
	    || fifo_buffer_is_empty (gXine.mBufVideo))) {
      
      xine_stop();
      
      break;
    }

    /*
     * make sure audio buffer is filled
     */
    
    audio_buf_status = (gAudioOut->get_buf_low) () ;
    
    if (audio_buf_status) {
      
      debug_printf (2,"xine_loop : audio low\n");
      
      if (audio_buf_status>1) {
	debug_printf (1, "xine_loop : audio emergency\n");
	
	/* mpeg2_set_skip_frame ();  */
      }
      
      
      if (!fifo_buffer_is_empty (gXine.mBufAudio)) {
	debug_printf (2,"xine_loop : decode audio\n"); 
	
	xine_decode_audio ();
      }	  
    }
    
    /*
     * decode video if we got time to do that
     */
    
    audio_buf_status = (gAudioOut->get_buf_low) () ;
    video_buf_full = vo_is_buffer_full ();
    
    if ( ((audio_buf_status>1) || !video_buf_full )
	 && !fifo_buffer_is_empty (gXine.mBufVideo) ) {
      
      debug_printf (2, "xine_loop : decode video (audio buf status: %d, video buf full:%d)\n",audio_buf_status,video_buf_full); 
      
      xine_decode_video ();
      
    } 
#ifdef VERBOSITY_LEVEL
    else {
      
      if (vo_is_buffer_full ())
	debug_printf (2, "xine_loop : no video decoding because frame buffer is full\n");
      
      if (fifo_buffer_is_empty (gXine.mBufVideo)) 
	debug_printf (2, "xine loop : no video decoding because video fifo buffer is empty\n");
      
    }
#endif
    
    /*
     * display video
     */
    
    vo_video_out ();
    
  }

  gXine.mStatusCallback (gXine.mnStatus);

  printf ("xine loop finished.\n");

  return NULL;
}


void xine_stop () {
  void *p;

  if (!gXine.mCurInput) 
    return;

  gXine.mnStatus = XINE_STOP;

  gXine.mCurDemux->demux_stop ();

  pthread_join (gXine.mThread, &p);

  mpeg2_reset ();
  ac3_reset ();
  if (gXine.mpgaudio_initialized) {
    exitMPGAUDIO (&gXine.mpgaudio);
    gXine.mpgaudio_initialized = 0;
  }

  fifo_buffer_clear(gXine.mBufVideo);
  fifo_buffer_clear(gXine.mBufAudio);

  vo_reset ();
  gAudioOut->close ();

  /* vo_set_logo_mode (1); */

  metronom_reset();

  gXine.mCurInput->close();

}

void xine_play_internal (char *MRL, int spos, uint32_t pos) {

  double     share ;
  uint32_t   len;
  int        i;

  printf ("xine open %s, start pos = %d\n",MRL, spos);

  if (gXine.mnStatus == XINE_PAUSE) {
    xine_pause();
    return;
  }

  if (gXine.mnStatus != XINE_STOP) {
    xine_stop ();
  }

  /*
   * find demuxer
   */

  gXine.mCurDemux = NULL;
  for (i=0; i<gXine.mNumDemuxer; i++) {
    if (gXine.mDemuxer[i]->demux_can_handle (MRL)) {
      gXine.mCurDemux = gXine.mDemuxer[i];
      break;
    }
  }

  if (!gXine.mCurDemux) {
    printf ("error: couldn't find demuxer for >%s<\n",MRL);
    return;
  }

  /*
   * find input plugin
   */
   
  gXine.mCurInput = NULL;

  for (i = 0; i<gXine.mNumInputPlugins; i++) {
    if (gXine.mInputPlugins[i].open(MRL)) {
      gXine.mCurInput = &gXine.mInputPlugins[i];
      break;
    }
  }

  if (!gXine.mCurInput) {
    perror ("open input source");
    return;
  }
  
  vo_set_logo_mode (0);

  /*
   * metronom
   */

  metronom_reset();

  /*
   * start demuxer
   */

  if (spos) {
    len = gXine.mCurInput->get_length ();
    share = (double) spos / 65535;
    pos = (uint32_t) (share * (double) len) ;
  }

  gXine.mCurDemux->demux_select_audio_channel (gXine.mnAudioChannel);
  gXine.mCurDemux->demux_start (gXine.mCurInput, gXine.mBufVideo,
				gXine.mBufAudio, pos);

  gXine.mnStatus = XINE_PLAY;

  /*
   * remember MRL
   */

  strncpy (gXine.mCurMRL, MRL, 1024);

  /*
   * start xine_loop
   */

  pthread_create (&gXine.mThread, NULL, xine_loop, NULL) ;
}

void xine_play (char *MRL, int spos) {

  xine_play_internal (MRL, spos, 0);

}

void xine_exit () {
  gXine.mnStatus = XINE_QUIT;

  config_file_save ();
}

void xine_pause () {
  void *p;

  if (gXine.mnStatus == XINE_PAUSE) {

    printf ("xine play %s from %d\n",gXine.mCurMRL, gXine.mnPausePos);

    gXine.mnStatus = XINE_STOP;
    xine_play_internal (gXine.mCurMRL, 0, gXine.mnPausePos);
    gXine.mnPausePos = 0;

  } else if (gXine.mnStatus == XINE_PLAY) {

    if (!gXine.mCurInput) 
      return;

    gXine.mnStatus = XINE_PAUSE;

    gXine.mnPausePos = gXine.mCurInput->seek (0, SEEK_CUR) - 3490000 ;

    if (gXine.mnPausePos < 0)
      gXine.mnPausePos = 0;

    gXine.mCurDemux->demux_stop ();
    
    pthread_join (gXine.mThread, &p);
    
    mpeg2_reset ();
    ac3_reset ();
    if (gXine.mpgaudio_initialized) {
      exitMPGAUDIO (&gXine.mpgaudio);
      gXine.mpgaudio_initialized = 0;
    }

    fifo_buffer_clear(gXine.mBufVideo);
    fifo_buffer_clear(gXine.mBufAudio);
    
    vo_reset ();
    gAudioOut->close ();

    metronom_reset();

    gXine.mCurInput->close();
  }
}

void xine_load_input_plugins () {
  DIR *dir;
  
  gXine.mNumInputPlugins = 0;
  
  dir = opendir (XINE_PLUGINDIR) ;
  
  if (dir) {
    struct dirent *pEntry;
    
    while ((pEntry = readdir (dir)) != NULL) {
      
      char str[1024];
      void *plugin;
      
      int nLen = strlen (pEntry->d_name);
      
      if ((pEntry->d_name[nLen-3]=='.') 
	  && (pEntry->d_name[nLen-2]=='s')
	  && (pEntry->d_name[nLen-1]=='o')) {

	/*
	 * input plugin found => load it
	 */
	
	sprintf (str, "%s/%s", XINE_PLUGINDIR, pEntry->d_name);
	
	if(!(plugin = dlopen (str, RTLD_LAZY))) {
	  fprintf(stderr, "%s(%d): %s doesn't seem to be installed (%s)\n", 
		  __FILE__, __LINE__, str, dlerror());
	  exit(1);
	}
	else {
	  
	  if (plugin) {
	    void *(*getinfo) (void);
	    
	    if((getinfo = dlsym(plugin, "input_plugin_getinfo")) != NULL) {
	      input_plugin_t *ipp;
	      
	      ipp = (input_plugin_t *) getinfo();
	      ipp->handle = plugin;
	      ipp->filename = str;
	      gXine.mInputPlugins[gXine.mNumInputPlugins] = *ipp; 
	    }
	    
	    gXine.mInputPlugins[gXine.mNumInputPlugins].init();
	    
	    fprintf(stderr, "input plugin found : %s(%s)\n", 
		    gXine.mInputPlugins[gXine.mNumInputPlugins].filename,
		    pEntry->d_name);
	    
	    gXine.mNumInputPlugins++;

	    if(gXine.mNumInputPlugins > INPUT_PLUGIN_MAX) {
	      fprintf(stderr, "%s(%d): too many input plugins installed, "
		      "exiting.\n", __FILE__, __LINE__);
	      exit(1);
	    }

	  }
	}

      }
    }
  }

  if (gXine.mNumInputPlugins == 0) {
    printf ("No input plugins found in %s! - "
	    "Did you install xine correctly??\n", XINE_PLUGINDIR);
    exit (1);
  }

}

void xine_init (int no_xv, gui_status_callback_func_t gui_status_callback) {

  ac3_config_t    ac3_config;
  ImlibImage     *img;
  char            strLogo[256];

  gXine.mStatusCallback = gui_status_callback;

  /*
   * init imlib
   */

  gImlib_data   = Imlib_init(gDisplay);

  /*
   * load logo (IMPORTANT ;))
   */
  sprintf (strLogo, "%s/xine_logo.png", XINE_SKINDIR);
  img = Imlib_load_image(gImlib_data, strLogo);

  if (!img) {
    fprintf (stderr, "logo not found :-((\n");
    exit(1);
  }

  Imlib_render (gImlib_data, img, img->rgb_width, img->rgb_height);
  gXineLogo  = Imlib_move_image(gImlib_data, img);
  gXineLogoWidth  = img->rgb_width ;
  gXineLogoHeight = img->rgb_height;

  /*
   * init video output
   */

  vo_init (no_xv);
  mpeg2_init();

  /*
   * init audio output
   */

  ac3_config.num_output_ch = 2;
  ac3_config.flags = 0;

  if((gAudioOut = audio_out_init(ao_use_driver)) == NULL) {
    fprintf(stderr, "audio driver '%s' failed, using null driver instead.\n", 
	    ao_available[ao_use_driver]);
    ao_use_driver = AO_NULL;
    gAudioOut = audio_out_init(AO_NULL);
  }

  ac3_init(&ac3_config, gAudioOut);
  gXine.mpgaudio_initialized = 0;

  /*
   * load input plugins
   */

  xine_load_input_plugins ();

  /*
   * init demuxer
   */
  
  gXine.mDemuxer[0] = init_demux_mpeg ();
  gXine.mDemuxer[1] = init_demux_mpeg_block ();
  gXine.mNumDemuxer = 2;
  gXine.mCurDemux = NULL;
  gXine.mnAudioChannel = 0;

  /*
   * Init buffers
   */

  buffer_pool_init (1400, 4096);
  gXine.mBufVideo = fifo_buffer_new ();
  gXine.mBufAudio = fifo_buffer_new ();
}

void xine_select_audio_channel (int nChannel) {

  gXine.mnAudioChannel = nChannel;

  if (gXine.mCurDemux) {
    gXine.mCurDemux->demux_select_audio_channel (gXine.mnAudioChannel);
  }
}

input_plugin_t* xine_get_input_plugin_list (int *nInputPlugins) {

  *nInputPlugins = gXine.mNumInputPlugins;
  return gXine.mInputPlugins;
}

int xine_get_current_position () {

  uint32_t pos, len;
  double   share;

  if (!gXine.mCurInput) {
    printf ("xine_get_current_position: no input source\n");
    return 0;
  }

  pos = gXine.mCurInput->seek (0, SEEK_CUR);

  len = gXine.mCurInput->get_length ();

  share = (double) pos / (double) len * 65535;

  return (int) share;
}


